from unittest.mock import patch
import pcleaner.gui.log_parser as lp
from tests.helpers import read_mock_file


def test_parse_good():
    # Test with good_pcleaner.log
    logfile = read_mock_file("good_pcleaner.log")
    # mock the get_username function to return "testvm"
    with patch("pcleaner.gui.log_parser.get_username", return_value="testvm"):
        sessions = lp.parse_log_file(logfile)

    assert len(sessions) == lp.MAX_SESSIONS
    assert sessions[19].criticals == 0
    assert sessions[19].errors == 0
    assert sessions[1].criticals == 3
    assert sessions[1].errors == 0
    assert sessions[1].corrupted is False



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ErrorDialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QHBoxLayout, QLabel,
    QPushButton, QSizePolicy, QSpacerItem, QVBoxLayout,
    QWidget)

from pcleaner.gui.log_viewer import LogViewer

class Ui_ErrorDialog(object):
    def setupUi(self, ErrorDialog):
        if not ErrorDialog.objectName():
            ErrorDialog.setObjectName(u"ErrorDialog")
        ErrorDialog.resize(1200, 700)
        ErrorDialog.setWindowTitle(u"Dialog")
        self.horizontalLayout_2 = QHBoxLayout(ErrorDialog)
        self.horizontalLayout_2.setSpacing(12)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(-1, 12, -1, -1)
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.label_error_icon = QLabel(ErrorDialog)
        self.label_error_icon.setObjectName(u"label_error_icon")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_error_icon.sizePolicy().hasHeightForWidth())
        self.label_error_icon.setSizePolicy(sizePolicy)
        self.label_error_icon.setMinimumSize(QSize(64, 64))
        self.label_error_icon.setMaximumSize(QSize(64, 64))
        self.label_error_icon.setText(u"<error icon>")

        self.verticalLayout_2.addWidget(self.label_error_icon)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer)


        self.horizontalLayout_2.addLayout(self.verticalLayout_2)

        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label_message = QLabel(ErrorDialog)
        self.label_message.setObjectName(u"label_message")
        self.label_message.setText(u"<error message>")

        self.verticalLayout.addWidget(self.label_message)

        self.log_viewer = LogViewer(ErrorDialog)
        self.log_viewer.setObjectName(u"log_viewer")

        self.verticalLayout.addWidget(self.log_viewer)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label_name_hidden = QLabel(ErrorDialog)
        self.label_name_hidden.setObjectName(u"label_name_hidden")
        self.label_name_hidden.setText(u"<name was hidden>")

        self.horizontalLayout.addWidget(self.label_name_hidden)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.pushButton_clipboard = QPushButton(ErrorDialog)
        self.pushButton_clipboard.setObjectName(u"pushButton_clipboard")
        icon = QIcon()
        iconThemeName = u"edit-copy"
        if QIcon.hasThemeIcon(iconThemeName):
            icon = QIcon.fromTheme(iconThemeName)
        else:
            icon.addFile(u".", QSize(), QIcon.Normal, QIcon.Off)

        self.pushButton_clipboard.setIcon(icon)

        self.horizontalLayout.addWidget(self.pushButton_clipboard)

        self.pushButton_open_issues = QPushButton(ErrorDialog)
        self.pushButton_open_issues.setObjectName(u"pushButton_open_issues")
        icon1 = QIcon()
        iconThemeName = u"internet-services"
        if QIcon.hasThemeIcon(iconThemeName):
            icon1 = QIcon.fromTheme(iconThemeName)
        else:
            icon1.addFile(u".", QSize(), QIcon.Normal, QIcon.Off)

        self.pushButton_open_issues.setIcon(icon1)

        self.horizontalLayout.addWidget(self.pushButton_open_issues)

        self.pushButton_close = QPushButton(ErrorDialog)
        self.pushButton_close.setObjectName(u"pushButton_close")
        icon2 = QIcon()
        iconThemeName = u"window-close"
        if QIcon.hasThemeIcon(iconThemeName):
            icon2 = QIcon.fromTheme(iconThemeName)
        else:
            icon2.addFile(u".", QSize(), QIcon.Normal, QIcon.Off)

        self.pushButton_close.setIcon(icon2)

        self.horizontalLayout.addWidget(self.pushButton_close)


        self.verticalLayout.addLayout(self.horizontalLayout)


        self.horizontalLayout_2.addLayout(self.verticalLayout)


        self.retranslateUi(ErrorDialog)

        QMetaObject.connectSlotsByName(ErrorDialog)
    # setupUi

    def retranslateUi(self, ErrorDialog):
        self.pushButton_clipboard.setText(QCoreApplication.translate("ErrorDialog", u"Copy to Clipboard", None))
        self.pushButton_open_issues.setText(QCoreApplication.translate("ErrorDialog", u"Open Issue Tracker", None))
        self.pushButton_close.setText(QCoreApplication.translate("ErrorDialog", u"Close", None))
        pass
    # retranslateUi




import io
import re
import sys
from types import TracebackType
from io import StringIO
from pathlib import Path

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from loguru import logger

from pcleaner.gui.error_dialog_driver import ErrorDialog
import pcleaner.gui.worker_thread as wt
from pcleaner.helpers import tr


# For all show functions, pad the dialog message, so that the dialog is not too narrow for the window title.
MIN_MSG_LENGTH = 50


class SelectableMessageBox(Qw.QMessageBox):
    """
    Subclass the QMessageBox to make the text selectable.
    """

    def __init__(self, *args, **kwargs):
        super(SelectableMessageBox, self).__init__(*args, **kwargs)
        self.setTextInteractionFlags(Qc.Qt.TextSelectableByMouse)

        labels = self.findChildren(Qw.QLabel)
        for label in labels:
            label.setTextInteractionFlags(Qc.Qt.TextSelectableByMouse)


def show_exception(
    parent,
    title: str,
    msg: str,
    worker_error: None | wt.WorkerError = None,
    collect_exception: bool = True,
) -> None:
    """
    Show an exception in a dialog along with logs.

    :param parent: The parent widget.
    :param title: The title of the dialog.
    :param msg: The message to show.
    :param worker_error: [Optional] A worker error object to read the exception from.
    :param collect_exception: [Optional] Whether to add exception information to the log.
    """

    if collect_exception:
        exception_type: type[BaseException]
        exception_value: BaseException
        exception_traceback: TracebackType

        if worker_error is not None:
            exception_type = worker_error.exception_type
            exception_value = worker_error.value
            exception_traceback = worker_error.traceback
        else:
            exception_type, exception_value, exception_traceback = sys.exc_info()

        logger.opt(
            depth=1, exception=(exception_type, exception_value, exception_traceback)
        ).critical(msg)

    box = ErrorDialog(parent, title, msg)
    box.exec()


def show_critical(parent, title: str, msg: str, **kwargs) -> int:
    msg = msg.ljust(MIN_MSG_LENGTH)
    box = SelectableMessageBox(
        Qw.QMessageBox.Critical,
        title,
        msg,
        Qw.QMessageBox.Yes | Qw.QMessageBox.Abort,
        parent,
        **kwargs,
    )
    return box.exec()


def show_warning(parent, title: str, msg: str) -> None:
    msg = msg.ljust(MIN_MSG_LENGTH)
    box = SelectableMessageBox(Qw.QMessageBox.Warning, title, msg, Qw.QMessageBox.Ok, parent)
    box.exec()


def show_info(parent, title: str, msg: str) -> None:
    msg = msg.ljust(MIN_MSG_LENGTH)
    box = SelectableMessageBox(Qw.QMessageBox.Information, title, msg, Qw.QMessageBox.Ok, parent)
    box.exec()


def show_question(
    parent, title: str, msg: str, buttons=Qw.QMessageBox.Yes | Qw.QMessageBox.Cancel
) -> int:
    msg = msg.ljust(MIN_MSG_LENGTH)
    dlg = Qw.QMessageBox(parent)
    dlg.setWindowTitle(title)
    dlg.setText(msg)
    dlg.setStandardButtons(buttons)
    dlg.setIcon(Qw.QMessageBox.Question)
    return dlg.exec()


def open_file(path: Path) -> None:
    """
    Open any given file with the default application.
    """
    logger.debug(f"Opening file {path}")
    try:
        # Use Qt to open the file, so that it works on all platforms.
        Qg.QDesktopServices.openUrl(Qc.QUrl.fromLocalFile(str(path)))
    except:
        show_exception(None, tr("File Error"), tr("Failed to open file."))


class CaptureOutput(Qc.QObject):
    text_written = Qc.Signal(str, str)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._io_stdout = io.StringIO()
        self._io_stderr = io.StringIO()

    def __enter__(self) -> "CaptureOutput":
        self._original_stdout = sys.stdout
        self._original_stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        sys.stdout = self._original_stdout
        sys.stderr = self._original_stderr

    def write(self, string, stream="stdout") -> None:
        if stream == "stdout":
            self._io_stdout.write(string)
        elif stream == "stderr":
            self._io_stderr.write(string)
        self.text_written.emit(string, stream)

    def getvalue(self, stream="stdout") -> str:
        if stream == "stdout":
            return self._io_stdout.getvalue()
        elif stream == "stderr":
            return self._io_stderr.getvalue()

    def flush(self) -> None:
        self._io_stdout.flush()
        self._io_stderr.flush()


# Mapping between KDE color keys and QPalette roles
section_role_mapping = {
    "Colors:Button": {
        "BackgroundNormal": Qg.QPalette.Button,
        "ForegroundNormal": Qg.QPalette.ButtonText,
        "ForegroundActive": Qg.QPalette.BrightText,
    },
    "Colors:View": {
        "BackgroundNormal": Qg.QPalette.Base,
        "BackgroundAlternate": Qg.QPalette.AlternateBase,
        "ForegroundNormal": Qg.QPalette.Text,
        "ForegroundLink": Qg.QPalette.Link,
        "ForegroundVisited": Qg.QPalette.LinkVisited,
        "ForegroundInactive": Qg.QPalette.PlaceholderText,
    },
    "Colors:Selection": {
        "BackgroundNormal": Qg.QPalette.Highlight,
        "ForegroundNormal": Qg.QPalette.HighlightedText,
    },
    "Colors:Tooltip": {
        "BackgroundNormal": Qg.QPalette.ToolTipBase,
        "ForegroundNormal": Qg.QPalette.ToolTipText,
    },
    "Colors:Window": {
        "BackgroundNormal": Qg.QPalette.Window,
        "ForegroundNormal": Qg.QPalette.WindowText,
    },
    # Add more mappings as needed
}


def clamp_8bit(value: int) -> int:
    """
    Clamp a value to the 0-255 range.
    """
    return max(0, min(value, 255))


def apply_color_effect(source: Qg.QColor, effect_base: Qg.QColor, contrast_amount: float) -> Qg.QColor:
    """
    Apply a color effect to a source color.

    :param source: The source color.
    :param effect_base: The base color for the effect.
    :param contrast_amount: The amount of contrast to apply.
    :return: The modified color.
    """
    # Essentially alpha blend, with the effect having the contrast amount as the alpha pasted on top.
    r = clamp_8bit(int(source.red() * (1 - contrast_amount) + effect_base.red() * contrast_amount))
    g = clamp_8bit(int(source.green() * (1 - contrast_amount) + effect_base.green() * contrast_amount))
    b = clamp_8bit(int(source.blue() * (1 - contrast_amount) + effect_base.blue() * contrast_amount))
    return Qg.QColor(r, g, b)

def load_color_palette(theme: str) -> Qg.QPalette:
    """
    Provide a theme name and get a QPalette object.
    The name should match one of the files in the themes folder.

    :param theme: The name of the theme.
    :return: A QPalette object.
    """
    palette = Qg.QPalette()

    file = Qc.QFile(f":/themes/{theme}")
    if file.open(Qc.QFile.ReadOnly | Qc.QFile.Text):
        stream = Qc.QTextStream(file)
        content = stream.readAll()

        # Find the disabled color parameters.
        disabled_color = Qg.QColor.fromRgb(128, 128, 128)  # Default to gray.
        disabled_contrast_amount = 0.0
        in_disabled_section = False
        for line in content.split("\n"):
            line = line.strip()
            if line == "[ColorEffects:Disabled]":
                in_disabled_section = True
                continue
            if not in_disabled_section:
                continue
            if line.startswith("["):
                in_disabled_section = False
                break
            if "=" not in line:
                continue

            # Ok, now we're in the disabled section.
            key, value = map(str.strip, line.split("=", 1))
            if key == "Color":
                r, g, b = map(int, value.split(","))
                disabled_color.setRgb(r, g, b)
            elif key == "ContrastAmount":
                disabled_contrast_amount = float(value)

        section = None
        for line in content.split("\n"):
            line = line.strip()
            if line.startswith("[") and line.endswith("]"):
                section = line[1:-1]
            elif "=" in line:
                key, value = map(str.strip, line.split("=", 1))
                role_mapping = section_role_mapping.get(section, {})
                qt_color_role = role_mapping.get(key, None)
                if qt_color_role is not None:
                    r, g, b = map(int, value.split(","))
                    palette.setColor(Qg.QPalette.Normal, qt_color_role, Qg.QColor(r, g, b))
                    palette.setColor(Qg.QPalette.Inactive, qt_color_role, Qg.QColor(r, g, b))
                    # Calculate the disabled color.
                    disabled_color = apply_color_effect(Qg.QColor(r, g, b), disabled_color, disabled_contrast_amount)
                    palette.setColor(Qg.QPalette.Disabled, qt_color_role, disabled_color)
    else:
        raise ValueError(f"Could not open theme file: {theme}")

    # Fallback calculations
    if not palette.color(Qg.QPalette.Light).isValid():
        base = palette.color(Qg.QPalette.Button)
        palette.setColor(Qg.QPalette.Light, base.lighter(150))

    if not palette.color(Qg.QPalette.Dark).isValid():
        base = palette.color(Qg.QPalette.Window)
        palette.setColor(Qg.QPalette.Dark, base.darker(150))

    if not palette.color(Qg.QPalette.Midlight).isValid():
        light = palette.color(Qg.QPalette.Light)
        button = palette.color(Qg.QPalette.Button)
        midlight = Qg.QColor(
            (light.red() + button.red()) // 2,
            (light.green() + button.green()) // 2,
            (light.blue() + button.blue()) // 2,
        )
        palette.setColor(Qg.QPalette.Midlight, midlight)

    if not palette.color(Qg.QPalette.Mid).isValid():
        dark = palette.color(Qg.QPalette.Dark)
        button = palette.color(Qg.QPalette.Button)
        mid = Qg.QColor(
            (dark.red() + button.red()) // 2,
            (dark.green() + button.green()) // 2,
            (dark.blue() + button.blue()) // 2,
        )
        palette.setColor(Qg.QPalette.Mid, mid)

    if not palette.color(Qg.QPalette.Shadow).isValid():
        palette.setColor(Qg.QPalette.Shadow, Qg.QColor(0, 0, 0))

    return palette


def ansi_to_html(input_text: str) -> str:
    """
    Convert ansi escape sequences to html.

    Analytics HTML template
    <html><head><meta name="qrichtext" content="1" /><style type="text/css">
    p, li { white-space: pre-wrap; font-family: Noto Mono, Monospace; size: 12pt}
    </style></head><body>
    <p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px;
     -qt-block-indent:0; text-indent:0px;"><span style=" ">Mask Fitment Analytics
     <br />---------------------- <br />Total boxes: 5 | Masks succeeded: 5 (100%)
     | Masks failed: </span><span style="color:#b21818;">0</span><span style=" ">
     <br />Perfect masks: </span><span style="color:#18b2b2;">5</span><span style=" ">
     (100%) | Average border deviation: 0.00 <br /><br />Mask usage by mask size
     (smallest to largest): <br />Mask 0  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 1  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 2  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 3  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 4  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 5  :  </span><span style="color:#18b2b2;">0</span>
     <span style=" "> / 0 <br />Mask 6  : </span><span style="color:#18b2b2;">
     ████████████████████████████████████████</span><span style=" "> </span>
     <span style="color:#18b2b2;">1</span><span style=" "> / 1 <br />Mask 7  :  </span>
     <span style="color:#18b2b2;">0</span><span style=" "> / 0 <br />Mask 8  :  </span>
     <span style="color:#18b2b2;">0</span><span style=" "> / 0 <br />Mask 9  :  </span>
     <span style="color:#18b2b2;">0</span><span style=" "> / 0 <br />Mask 10 :  </span>
     <span style="color:#18b2b2;">0</span><span style=" "> / 0 <br />Box mask: </span>
     <span style="color:#18b2b2;">████████████████████████████████████████</span>
     <span style=" "> </span><span style="color:#18b2b2;">4</span><span style=" ">
     / 4 <br /><br /></span><span style="color:#18b2b2;">█ Perfect</span><span style=" ">
     | █ Total<br /><br /></span></p></body></html>
    """
    # Define ANSI escape sequences and corresponding Qt HTML color codes
    ansi_to_html_lut = {
        "31": "color:#eb514f;",
        "32": "color:#00ff00;",
        "92": "color:#22ff22;",
        "33": "color:#ffff00;",
        "34": "color:#0000ff;",
        "35": "color:#a771bf;",
        "36": "color:#00a3a3;",
        "37": "color:#ffffff;",
        "39": " ",  # Default color
    }

    # Initialize HTML header and footer
    html_header = """<html><head><meta name="qrichtext" content="1" /><style type="text/css">
        p, li { white-space: pre-wrap; font-family: Noto Mono, Monospace; size: 12pt}
        </style></head><body>
        <p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;">"""
    html_footer = "</p></body></html>"

    # Initialize output HTML with header
    buffer = StringIO()
    buffer.write(html_header)

    # Initialize current_style
    current_style = ""

    # Strip leading and trailing whitespace.
    input_text = input_text.strip()

    # Use regex to find all ANSI sequences and split text
    for plain_text, ansi_code in re.findall(r"(.*?)(?:\x1b\[(\d+)m|$)", input_text, re.DOTALL):
        # Add plain text
        newline = "\n"
        buffer.write(
            f'<span style="{current_style}">{plain_text.replace(newline, "<br />")}</span>'
        )

        # Update the style for the next span based on ANSI code
        if ansi_code:
            current_style = ansi_to_html_lut.get(ansi_code, "")

    # Add HTML footer
    buffer.write(html_footer)

    return buffer.getvalue()



import itertools
import shutil
from collections import defaultdict
from io import StringIO
from pathlib import Path
from typing import Sequence

import colorama as clr
from attrs import frozen
from natsort import natsorted

import pcleaner.helpers as hp
from pcleaner.helpers import tr
import pcleaner.structures as st
import pcleaner.config as cfg


@frozen
class OCRAnalytic:
    num_boxes: int
    small_box_sizes: Sequence[int]
    removed_box_sizes: Sequence[int]


# I'll admit this file is a mess.
# These bar chart creators should be abstracted into a single implementation, not 3 different ones.


def terminal_width() -> int:
    # Use a fallback of 100 for unsupported terminals.
    width = shutil.get_terminal_size((100, 100)).columns
    # Make sure the width is at least 20, to avoid weirdness.
    return max(width, 20)


def show_ocr_analytics(
    analytics: Sequence[st.OCRAnalytic],
    max_ocr_size: int,
    max_columns: int = 100,
) -> str:
    """
    The analytics are gathered from the OCR pre-processing which has the goal of
    finding small boxes and to check if they contain text worth cleaning.

    The analytics for each page in the list consist of the following:
    - The number of boxes on the page.
    - The sizes of the small boxes.
    - The sizes of the boxes that were removed.
    - list of: The original file path and the removed text and the box it belonged to.

    Various statistics are then shown.

    :param analytics: The analytics gathered from each pre-processed page.
    :param max_ocr_size: The maximum size of a box to be considered small.
    :param max_columns: The maximum number of columns to use for the chart per line.
    :return: The analytics as a string.
    """
    buffer = StringIO()

    num_boxes = sum(a.num_boxes for a in analytics)
    num_small_boxes = sum(len(a.box_sizes_ocr) for a in analytics)
    small_box_sizes = list(itertools.chain.from_iterable(a.box_sizes_ocr for a in analytics))
    removed_box_sizes = list(itertools.chain.from_iterable(a.box_sizes_removed for a in analytics))
    small_box_ratio = f"{num_small_boxes / num_boxes:.0%}" if num_boxes > 0 else tr("N/A")
    removed_box_ratio = f"{len(removed_box_sizes) / num_boxes:.0%}" if num_boxes > 0 else tr("N/A")
    removed_among_small_ratio = (
        f"{len(removed_box_sizes) / num_small_boxes:.0%}" if num_small_boxes > 0 else tr("N/A")
    )
    # Partition the small box sizes into 5 rough size ranges, equally spaced.
    part_size = 500 if max_ocr_size < 500 * 10 else max_ocr_size // 10
    partitioned_small_box_sizes = partition_list(
        small_box_sizes, partition_size=part_size, max_value=max_ocr_size
    )
    partitioned_removed_boxes = partition_list(
        removed_box_sizes, partition_size=part_size, max_value=max_ocr_size
    )
    title = tr("OCR Analytics")
    buffer.write(f"\n{title}\n")
    buffer.write(f"{'━' * len(title)}\n")
    buffer.write(
        tr(
            "Number of boxes: {num_boxes} | "
            "Number of small boxes: {num_small_boxes} ({small_box_ratio})\n"
        ).format(
            num_boxes=num_boxes,
            num_small_boxes=num_small_boxes,
            small_box_ratio=small_box_ratio,
        )
    )
    buffer.write(
        tr(
            "Number of removed boxes: {num_removed} ({removed_box_ratio} total, {removed_among_small_ratio} of small boxes)\n"
        ).format(
            num_removed=len(removed_box_sizes),
            removed_box_ratio=removed_box_ratio,
            removed_among_small_ratio=removed_among_small_ratio,
        )
    )
    if small_box_sizes:
        buffer.write(tr("\nSmall box sizes:\n"))
        buffer.write(
            draw_pretty_ocr_result_chart(
                partitioned_small_box_sizes, partitioned_removed_boxes, max_columns
            )
            + "\n"
        )
    else:
        buffer.write(tr("No not-removed small boxes found.\n"))

    # Show removed texts.
    removed_path_texts: list[tuple[Path, str, st.Box]] = list(
        itertools.chain.from_iterable(a.removed_box_data for a in analytics)
    )

    if removed_path_texts:
        # Find and then remove the longest common prefix from the file paths.
        paths, texts, _ = zip(*removed_path_texts)
        removed_paths = hp.trim_prefix_from_paths(paths)
        removed_path_texts = list(zip(removed_paths, texts))
        # Sort by file path.
        removed_path_texts = natsorted(removed_path_texts, key=lambda p: p[0])

    if removed_path_texts:
        buffer.write(tr("\nRemoved bubbles:\n"))
        for path, text in removed_path_texts:
            buffer.write(tr("Page {path}: {text}\n").format(path=path, text=text))

    buffer.write("\n")

    return buffer.getvalue()


def draw_pretty_ocr_result_chart(
    small_box_data: list[tuple[str, int]],
    removed_box_data: list[tuple[str, int]],
    max_columns: int = 100,
) -> str:
    """
    Draw a pretty stacked bar chart.

    :param small_box_data: The data for the small boxes.
    :param removed_box_data: The data for the removed boxes.
    :param max_columns: The maximum number of columns to use for the chart per line.
    :return: The chart as a string.
    """
    buffer = StringIO()

    data_array = [
        (label, lower, upper)
        for (label, lower), (_, upper) in zip(small_box_data, removed_box_data)
    ]
    max_label_width = max(len(label) for label, _, _ in data_array)
    bar_width = (
        max_columns - max_label_width - 25
    )  # 16 for the spacing and trailing number and padding.
    max_value = max(lower + upper for _, lower, upper in data_array)

    for label, lower, upper in data_array:
        lower_bar = "█" * int((lower - upper) / max_value * bar_width)
        upper_bar = "█" * int(upper / max_value * bar_width)
        buffer.write(
            f"{label:<{max_label_width}}: "
            f"{lower_bar}{clr.Fore.RED}{upper_bar} {upper}{clr.Style.RESET_ALL} / {lower}\n"
        )

    # Show legend.
    buffer.write(
        f"\n█ "
        + tr("Small boxes")
        + f" | {clr.Fore.RED}█ "
        + tr("Removed boxes")
        + f"{clr.Style.RESET_ALL}"
    )

    return buffer.getvalue()


def draw_masker_histogram(data: dict[str, tuple[int, int]], max_columns: int = 100) -> str:
    """
    Draw a histogram of the masker results.

    :param data: A dict of the mask name, (perfect uses, all uses).
    :param max_columns: The maximum number of columns to use for the histogram per line.
    :return: The histogram as a string.
    """

    # Find the maximum value in the data. (at least 1 to avoid division by zero)
    max_value = max(list(map(lambda x: x[1], data.values())))
    max_value = max(max_value, 1)

    # Longest key.
    max_label_width = max(len(k) for k in data.keys())
    bar_width = (
        max_columns - max_label_width - 25
    )  # 16 for the spacing and trailing number and padding.

    buffer = StringIO()

    # Draw the histogram.
    for key, value in data.items():
        # Use a Unicode character for the bar.
        left_bar = "█" * int(value[0] / max_value * bar_width)
        right_bar = "█" * int((value[1] - value[0]) / max_value * bar_width)
        # If both are 0 in length, but one of them isn't 0 in value, draw the right-most one as a sliver.
        if (not right_bar) and value[1] > value[0]:
            right_bar = "▏"
        if (not left_bar) and (not right_bar) and value[0] > 0:
            left_bar = "▏"

        if not key == tr("Failed"):
            buffer.write(
                f"{key:<{max_label_width}}: "
                f"{clr.Fore.CYAN}{left_bar}{clr.Style.RESET_ALL}{right_bar} "
                f"{clr.Fore.CYAN}{value[0]}{clr.Fore.RESET} / {value[1]}\n"
            )
        else:
            buffer.write(
                f"{key:<{max_label_width}}: "
                f"{clr.Fore.RED}{right_bar} {value[1]}{clr.Fore.RESET}\n"
            )

    # Show legend.
    buffer.write(
        f"\n{clr.Fore.CYAN}█ "
        + tr("Perfect")
        + f"{clr.Style.RESET_ALL} | █ "
        + tr("Total")
        + f" | {clr.Fore.RED}█ "
        + tr("Failed")
        + f"{clr.Style.RESET_ALL} \n"
    )

    return buffer.getvalue()


def partition_list(
    data: list[int],
    *,
    partition_size: int = 500,
    min_value: int | None = 0,
    max_value: int | None = None,
) -> list[tuple[str, int]]:
    """
    Partition the list into tuples of the form (range, count).
    The ranges are saved as a string: "0-499", "500-999", etc.
    If no max_value is given, the maximum value in the data is used.
    If min_value is None, the minimum value in the data is used. Otherwise it defaults to 0.

    :param data: The data to partition.
    :param partition_size: [Optional] The size of the partitions. Default: 500.
    :param min_value: [Optional] The minimum value to partition. Default: 0.
    :param max_value: [Optional] The maximum value to partition to. Default: None.
    :return: The partitioned data as a list of tuples.
    """
    if max_value is None:
        max_value = max(data)

    max_label_len = len(str(max_value))

    partition = [
        (
            f"{i * partition_size: >{max_label_len}}-{(i + 1) * partition_size: >{max_label_len}}",
            len([d for d in data if i * partition_size <= d < (i + 1) * partition_size]),
        )
        for i in range(min_value, max_value // partition_size)
    ]

    return partition


def show_masker_analytics(
    analytics: list[st.MaskFittingAnalytic], masker_config: cfg.MaskerConfig, max_columns: int = 100
) -> str:
    """
    Present the analytics gathered from the masking process.

    :param analytics: The analytics gathered from each page.
    :param masker_config: The masker configuration used to gather the analytics.
    :param max_columns: The maximum number of columns to use for the chart per line.
    :return: The analytics as a string.
    """
    buffer = StringIO()

    total_boxes = len(analytics)
    masks_succeeded = sum(1 for analytic in analytics if analytic.fit_was_found)
    perfect_masks = sum(
        1 for analytic in analytics if analytic.fit_was_found and analytic.mask_std_deviation == 0
    )
    average_border_deviation = (
        f"{sum(analytic.mask_std_deviation for analytic in analytics if analytic.fit_was_found) / masks_succeeded:.2f}"
        if masks_succeeded
        else tr("N/A")
    )
    success_rate = f"{masks_succeeded / total_boxes:.0%}" if total_boxes else tr("N/A")
    perfect_mask_rate = f"{perfect_masks / masks_succeeded:.0%}" if masks_succeeded else tr("N/A")
    masks_failed = total_boxes - masks_succeeded

    # Count how many times each mask thickness was used.
    mask_usage_by_thickness: defaultdict[int | None, int] = defaultdict(int)
    perfect_mask_usage_by_thickness: defaultdict[int | None, int] = defaultdict(int)
    for analytic in analytics:
        if analytic.fit_was_found:
            mask_usage_by_thickness[analytic.mask_thickness] += 1
            if analytic.mask_std_deviation == 0:
                perfect_mask_usage_by_thickness[analytic.mask_thickness] += 1

    # Write the analytics.
    title = tr("Mask Fitment Analytics")
    buffer.write(f"\n{title}\n")
    buffer.write(f"{'━' * len(title)}\n")
    buffer.write(
        tr("Total boxes")
        + f": {total_boxes} | "
        + tr("Masks succeeded")
        + f": {masks_succeeded} ({success_rate}) | "
        + tr("Masks failed")
        + f": {clr.Fore.RED}{masks_failed}{clr.Fore.RESET}\n"
    )
    buffer.write(
        tr("Perfect masks")
        + f": {clr.Fore.CYAN}{perfect_masks}{clr.Fore.RESET} ({perfect_mask_rate}) | "
        + tr("Average border deviation")
        + f": {average_border_deviation}\n"
    )
    buffer.write(tr("\nMask usage by thickness (in pixels):\n"))

    # Quit being retarded over mixed types in a list, this is Python for fuck's sake.
    # noinspection PyTypeChecker
    thicknesses: list[int | None] = list(
        range(
            masker_config.min_mask_thickness,
            masker_config.min_mask_thickness
            + masker_config.mask_growth_step_pixels * masker_config.mask_growth_steps,
            masker_config.mask_growth_step_pixels,
        )
    ) + [None]
    # Prepare histogram data: label, (perfect, total)
    mask_usages_dict: dict[str, tuple[int, int]] = {}
    for thickness in thicknesses:
        if thickness is not None:
            label = tr("Mask ({thickness}px)").format(thickness=thickness)
        else:
            label = tr("Box mask")

        perfect = perfect_mask_usage_by_thickness[thickness]
        total = mask_usage_by_thickness[thickness]

        mask_usages_dict[label] = (perfect, total)

    # Add failed masks to the histogram.
    mask_usages_dict[tr("Failed")] = (0, masks_failed)

    buffer.write(draw_masker_histogram(mask_usages_dict, max_columns) + "\n")

    pages_with_success_and_fails_dict: defaultdict[Path, defaultdict[str, int]] = defaultdict(
        lambda: defaultdict(int)
    )
    for analytic in analytics:
        pages_with_success_and_fails_dict[analytic.image_path][
            "succeeded" if analytic.fit_was_found else "failed"
        ] += 1

    if not pages_with_success_and_fails_dict:
        buffer.write(tr("All bubbles were successfully masked.\n"))
        return buffer.getvalue()

    pages_with_success_and_fails: list[tuple[Path, int, int]] = [
        (page_path, counts["succeeded"], counts["failed"])
        for page_path, counts in pages_with_success_and_fails_dict.items()
        if counts["failed"]
    ]

    if pages_with_success_and_fails:
        page_paths, succeeded_counts, failed_counts = zip(*pages_with_success_and_fails)
        page_paths = hp.trim_prefix_from_paths(page_paths)
        pages_with_success_and_fails = list(zip(page_paths, succeeded_counts, failed_counts))
        pages_with_success_and_fails = natsorted(pages_with_success_and_fails, key=lambda x: x[0])

    buffer.write(tr("\nPages with failures / total:\n"))
    for page_path, succeeded, failed in pages_with_success_and_fails:
        buffer.write(f"{page_path}: {clr.Fore.RED}{failed}{clr.Fore.RESET} / {succeeded+failed}\n")

    buffer.write("\n")

    return buffer.getvalue()


def show_denoise_analytics(
    analytics: list[st.DenoiseAnalytic],
    configured_deviation_threshold: float,
    max_deviation_threshold: float,
    max_columns: int = 100,
) -> str:
    """
    Present the analytics gathered from the denoising process.

    :param analytics: The analytics gathered from each page, here standard deviations.
    :param configured_deviation_threshold: The configured deviation threshold to de denoising.
    :param max_deviation_threshold: The maximum deviation threshold to discard the mask.
    :param max_columns: The maximum number of columns to use for the chart per line.
    :return: The analytics as a string.
    """
    buffer = StringIO()

    title = tr("Denoising Analytics")
    buffer.write(f"\n{title}\n")
    buffer.write(f"{'━' * len(title)}\n")

    # Get the standard deviations into one single list.
    std_deviations: list[float] = [
        *itertools.chain.from_iterable(analytic.std_deviations for analytic in analytics)
    ]

    # Split the standard deviations into two lists: one for the ones that were below the threshold,
    # and one for the ones that were above the threshold.
    below_threshold = [
        std_dev for std_dev in std_deviations if std_dev <= configured_deviation_threshold
    ]
    above_threshold = [
        std_dev for std_dev in std_deviations if std_dev > configured_deviation_threshold
    ]

    total_masks = len(std_deviations)
    above_threshold_count = len(above_threshold)
    above_ratio = f"{above_threshold_count / total_masks:.0%}" if total_masks else tr("N/A")

    buffer.write(
        tr("Total masks")
        + f": {total_masks} | "
        + tr("Masks denoised")
        + f": {clr.Fore.MAGENTA}{above_threshold_count}{clr.Fore.RESET} ({above_ratio})\n"
    )
    buffer.write(
        tr("Minimum deviation to denoise")
        + f": {configured_deviation_threshold} | "
        + tr("Maximum allowed deviation")
        + f": {max_deviation_threshold}\n"
    )
    buffer.write(tr("Standard deviation around masks:\n\n"))

    # Show the lists in a graph.
    buffer.write(
        draw_denoise_histogram(
            below_threshold,
            above_threshold,
            max_deviation_threshold,
            max_columns,
        )
        + "\n"
    )
    buffer.write("\n")

    return buffer.getvalue()


def draw_denoise_histogram(
    below_threshold: list[float],
    above_threshold: list[float],
    max_value: float,
    max_columns: int = 100,
) -> str:
    """
    Draw a histogram of the deviations.

    :param below_threshold: The standard deviations that were below the threshold.
    :param above_threshold: The standard deviations that were above the threshold.
    :param max_value: The maximum deviation threshold to discard the mask.
    :param max_columns: The maximum number of columns to use for the histogram per line.
    :return: The histogram as a string.
    """
    # Partition the values into 10 buckets.
    # Use a cubic polynomial that passes through 0 and the max value.
    # The polynomial is a simple f(x) = ax^3, meaning the coefficient a
    # must be max_value/10^3.
    exponent = 3
    cubic_coefficient = max_value / 10**exponent
    # The buckets contain: (min, max)
    bucket_ranges = (
        (cubic_coefficient * (i - 1) ** exponent, cubic_coefficient * i**exponent)
        for i in range(1, 11)
    )
    # Set the precision so that the second bucket isn't 0.
    # Basically, if the value is 0.00143, we want a precision of 3 to still see that 1.
    second_bucket_min = str(cubic_coefficient * 1**exponent).split(".")[1]
    # Count the zeroes from the left, and add 1 to only show up to the first non-zero digit.
    precision = len(second_bucket_min) - len(second_bucket_min.lstrip("0")) + 1
    # The buckets contain: (min, max, count_below_threshold, count_above_threshold)
    buckets: list[tuple[str, str, int, int]] = [
        (
            f"{b_min:.{precision}f}",
            f"{b_max:.{precision}f}",
            sum(1 for std_dev in below_threshold if b_min <= std_dev < b_max),
            sum(1 for std_dev in above_threshold if b_min <= std_dev < b_max),
        )
        for b_min, b_max in bucket_ranges
    ]
    # Format the ranges to look better.
    number_width = len(str(f"{max_value:.{precision}f}"))
    buckets_labeled: list[tuple[str, int, int]] = [
        (
            f"{b_min: >{number_width}}-{b_max: >{number_width}}",
            below,
            above,
        )
        for b_min, b_max, below, above in buckets
    ]
    # Get the maximum count in the buckets to rescale them.
    max_count = max((below + above for _, below, above in buckets_labeled))
    max_count = max(max_count, 1)
    # Rescale the buckets.

    max_label_length = max(len(label) for label, _, _ in buckets_labeled)
    bar_width = max_columns - max_label_length - 28  # 28 is for the spaces and the brackets.

    bucket_bar_lengths: list[tuple[int, int]] = [
        (int(below * bar_width / max_count), int(above * bar_width / max_count))
        for _, below, above in buckets_labeled
    ]

    buffer = StringIO()
    # Draw the buckets.
    for (b_range, below, above), (below_len, above_len) in zip(buckets_labeled, bucket_bar_lengths):
        left_bar = "█" * below_len
        right_bar = "█" * above_len
        if (not above_len) and above:
            right_bar = "▏"
        if (not right_bar) and (not left_bar) and below:
            left_bar = "▏"

        buffer.write(
            f"{b_range} : {left_bar}{clr.Fore.MAGENTA}{right_bar} {above}{clr.Fore.RESET} / {below+above}\n"
        )

    # Draw the legend.
    buffer.write(
        f"\n{clr.Fore.MAGENTA}█ " + tr("Denoised") + f"{clr.Fore.RESET} | █ " + tr("Total\n")
    )
    return buffer.getvalue()


def show_inpainting_analytics(
    analytics: list[st.InpaintingAnalytic],
    minimum_thickness: int,
    maximum_thickness: int,
    max_columns: int = 100,
) -> str:
    """
    Present the analytics gathered from the inpainting process.

    :param analytics: The analytics gathered from each page, here standard deviations.
    :param minimum_thickness: The minimum thickness of the inpainted outlines.
    :param maximum_thickness: The maximum thickness of the inpainted outlines.
    :param max_columns: The maximum number of columns to use for the chart per line.
    :return: The analytics as a string.
    """
    buffer = StringIO()

    title = tr("Inpainting Analytics")
    buffer.write(f"\n{title}\n")
    buffer.write(f"{'━' * len(title)}\n")

    # Get the thicknesses into one single list.
    thicknesses: list[int] = [
        *itertools.chain.from_iterable(analytic.thicknesses for analytic in analytics)
    ]

    total_inpaint = len(thicknesses)
    average_thickness = sum(thicknesses) / total_inpaint if total_inpaint else tr("N/A")

    buffer.write(
        tr("Inpainting performed")
        + f": {total_inpaint} | "
        + tr("Average thickness")
        + f": {average_thickness}\n"
    )
    buffer.write(
        tr("Minimum thickness")
        + f": {minimum_thickness} | "
        + tr("Maximum thickness")
        + f": {maximum_thickness}\n"
    )
    buffer.write(tr("Outline thickness around text inpainted:\n\n"))

    # Show the lists in a graph.
    buffer.write(
        draw_inpaint_histogram(
            thicknesses,
            minimum_thickness,
            maximum_thickness,
            max_columns,
        )
        + "\n"
    )
    buffer.write("\n")

    return buffer.getvalue()


def draw_inpaint_histogram(
    thicknesses: list[int],
    minimum_thickness: int,
    maximum_thickness: int,
    max_columns: int = 100,
) -> str:
    """
    Draw a histogram of the deviations.

    :param thicknesses: The thicknesses of the inpainted outlines.
    :param minimum_thickness: The minimum thickness of the inpainted outlines.
    :param maximum_thickness: The maximum thickness of the inpainted outlines.
    :param max_columns: The maximum number of columns to use for the histogram per line.
    :return: The histogram as a string.
    """
    # The buckets contain: (min, count)
    buckets: list[tuple[int, int]] = []
    for thickness in range(minimum_thickness, maximum_thickness + 1):
        buckets.append((thickness, thicknesses.count(thickness)))
    # Format the ranges to look better.
    number_width = len(str(buckets[-1][0]))
    buckets_labeled: list[tuple[str, int]] = [
        (
            f"{thiccness: >{number_width}}",
            count,
        )
        for thiccness, count in buckets
    ]
    # Get the maximum count in the buckets to rescale them.
    max_count = max((count for _, count in buckets_labeled))
    max_count = max(max_count, 1)
    # Rescale the buckets.

    max_label_length = max(len(label) for label, _ in buckets_labeled)
    bar_width = max_columns - max_label_length - 28  # 28 is for the spaces and the brackets.

    bucket_bar_lengths: list[int] = [
        int(count * bar_width / max_count) for _, count in buckets_labeled
    ]

    buffer = StringIO()
    # Draw the buckets.
    for (b_range, count), (count_len) in zip(buckets_labeled, bucket_bar_lengths):
        bar = "█" * count_len
        if (not count_len) and count:
            bar = "▏"

        buffer.write(f"{b_range} : {clr.Fore.LIGHTGREEN_EX}{bar} {count}{clr.Fore.RESET}\n")

    # Draw the legend.
    buffer.write(f"\n{clr.Fore.LIGHTGREEN_EX}█ " + tr("Inpainted") + clr.Fore.RESET + "\n")

    return buffer.getvalue()
